/*  o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
    o+
    o+     File:         express.cpp
    o+
    o+     Programmer:   Shaun Flisakowski
    o+     Date:         Aug 9, 1998
    o+
    o+     A high-level view of expressions.
    o+
    o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o  */

#include <string.h>

#include "express.h"

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(long val)
{
    etype = ET_Constant;
    ctype = CT_Int;
    lng = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(char val)
{
    etype = ET_Constant;
    ctype = CT_Char;
    ch = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(double val)
{
    etype = ET_Constant;
    ctype = CT_Float;
    doub = val;
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Constant::Constant(ConstantType ct, int val /* =0 */)
{
    etype = ET_Constant;
    ctype = ct;
    enm = val;
}
 
// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
StringConstant::StringConstant(char *str)
{
    etype = ET_Constant;
    ctype = CT_String;

    buff = new char[strlen(str)+1];
    strcpy(buff,str);
}

// o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o+o
Expression*
buildExpr(treenode *node)
{
    leafnode   *leaf;
    Expression *lExpr, *rExpr;
    Expression *result = NULL;

    lExpr = rExpr = NULL;

    if (node)
    {
        switch (node->which)
        {
            default:
            case NONE_T:
                break;
    
            case LEAF_T:
                leaf = (leafnode*) node;
                switch (leaf->tok)
                {
                    case CHAR_CONST:
                        result = new Constant(leaf->data.cval);
                        break;
    
                    case INUM:
                        result = new Constant(leaf->data.ival);
                        break;
    
                    case RNUM:
                        result = new Constant(leaf->data.dval);
                        break;
    
                    case STRING:
                        result = new StringConstant(leaf->data.str);
                        break;
                }
                break;
    
            case NODE_T:
                switch (node->type)
                {
                    case TN_EXPR:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);

                        switch (node->tok)
                        {
                            case PLUS:
                                if (lExpr)
                                  result = new BinaryExpr(BO_Plus,lExpr,rExpr);
                                else
                                  result = new UnaryExpr(UO_Plus,rExpr);
                                break;

                            case MINUS:
                                if (lExpr)
                                  result = new BinaryExpr(BO_Minus,lExpr,rExpr);
                                else
                                  result = new UnaryExpr(UO_Minus,rExpr);
                                break;

                            case STAR:
                                result = new BinaryExpr(BO_Mult,lExpr,rExpr);
                                break;

                            case DIV:
                                result = new BinaryExpr(BO_Div,lExpr,rExpr);
                                break;

                            case MOD:
                                result = new BinaryExpr(BO_Mod,lExpr,rExpr);
                                break;
            
                            case NOT:
                                result = new UnaryExpr(UO_Not,rExpr);
                                break;
            
                            case AND:
                                result = new BinaryExpr(BO_And,lExpr,rExpr);
                                break;

                            case OR:
                                result = new BinaryExpr(BO_Or,lExpr,rExpr);
                                break;
            
                            case B_NOT:
                                result = new UnaryExpr(UO_BitNot,rExpr);
                                break;

                            case B_AND:
                                result = new BinaryExpr(BO_BitAnd,lExpr,rExpr);
                                break;
            
                            case B_OR:
                                result = new BinaryExpr(BO_BitOr,lExpr,rExpr);
                                break;
            
                            case B_XOR:
                                result = new BinaryExpr(BO_BitXor,lExpr,rExpr);
                                break;
            
            
                            case L_SHIFT:
                                result = new BinaryExpr(BO_Shl,lExpr,rExpr);
                                break;
            
                            case R_SHIFT:
                                result = new BinaryExpr(BO_Shr,lExpr,rExpr);
                                break;
            
                            case INCR:          // ++
                                if (rExpr)
                                  result = new UnaryExpr(UO_PreInc,rExpr);
                                else
                                  result = new UnaryExpr(UO_PostInc,lExpr);
                                break;

                            case DECR:          // --
                                if (rExpr)
                                  result = new UnaryExpr(UO_PreDec,rExpr);
                                else
                                  result = new UnaryExpr(UO_PostDec,lExpr);
                                break;

                            // Relational Ops
                            case EQUAL:         // ==
                                result = new RelExpr(RO_Equal,lExpr,rExpr);
                                break;

                            case NOT_EQ:
                                result = new RelExpr(RO_NotEqual,lExpr,rExpr);
                                break;

                            case LESS:
                                result = new RelExpr(RO_Less,lExpr,rExpr);
                                break;

                            case LESS_EQ:
                                result = new RelExpr(RO_LessEql,lExpr,rExpr);
                                break;

                            case GRTR:
                                result = new RelExpr(RO_Grtr,lExpr,rExpr);
                                break;

                            case GRTR_EQ:
                                result = new RelExpr(RO_GrtrEql,lExpr,rExpr);
                                break;
                        }
                        break;
    
                    case TN_ASSIGN:
                        lExpr = buildExpr(node->lnode);
                        rExpr = buildExpr(node->rnode);

                        switch (node->tok)
                        {
                            case EQ:            // =
                                result = new AssignExpr(AO_Equal,lExpr,rExpr);
                                break;

                            case PLUS_EQ:
                                result = new AssignExpr(AO_PlusEql,lExpr,rExpr);
                                break;

                            case MINUS_EQ:
                                result = new AssignExpr(AO_MinusEql,lExpr,rExpr);
                                break;

                            case STAR_EQ:
                                result = new AssignExpr(AO_MultEql,lExpr,rExpr);
                                break;

                            case DIV_EQ:
                                result = new AssignExpr(AO_DivEql,lExpr,rExpr);
                                break;

                            case MOD_EQ:
                                result = new AssignExpr(AO_ModEql,lExpr,rExpr);
                                break;
            
                            case B_NOT_EQ:
                                result = new AssignExpr(AO_BitNotEql,lExpr,rExpr);
                                break;
            
                            case B_AND_EQ:
                                result = new AssignExpr(AO_BitAndEql,lExpr,rExpr);
                                break;
            
                            case B_OR_EQ:
                                result = new AssignExpr(AO_BitOrEql,lExpr,rExpr);
                                break;
            
                            case B_XOR_EQ:
                                result = new AssignExpr(AO_BitXorEql,lExpr,rExpr);
                                break;
            
            
                            case L_SHIFT_EQ:
                                result = new AssignExpr(AO_ShlEql,lExpr,rExpr);
                                break;
            
                            case R_SHIFT_EQ:
                                result = new AssignExpr(AO_ShrEql,lExpr,rExpr);
                                break;
                        }
                        break;
    
                    case TN_CAST:
                    case TN_DEREF:
                    case TN_INDEX:
                    case TN_EXPR_LIST:
                    case TN_SELECT:
                    case TN_FUNC_CALL:
                        break;
                }
                break;
    
            case IF_T:
                break;
            
            case FOR_T:
                break;
        }
    }

    return result;
}
