
/*  ###############################################################
    ##
    ##     C Tree Builder
    ##
    ##     File:         transtree.c
    ##
    ##     Programmer:   Kunle Olukotun and Mark Willey
    ##     Date:         June 3 1999
    ##
    ##
    ###############################################################  */

#include    <string.h>
#include    "prnttree.h"
#include    "transtree.h"
#include    "token.h"
#include    "gram.h"
#include    "symtab.h"
#include    "globals.h"


/* willey
needed to send code snippets to different files
*/
FILE *header_fp;
FILE *startup_fp;
FILE *shutdown_fp;
FILE *decl_fp;

static symentry_t *TrueLocals, *GlobalLocals;
static char *CurFunc, LoopName[256];
static int LoopNum;

/*  ###############################################################  */


static if_node  *trans_if(if_node *ifn);
static for_node *trans_for(for_node *forn);
static treenode *trans_while(treenode *whilen);
static treenode *trans_any(treenode *child);
static treenode *trans_node(treenode *node);

static state_t analyze_any(treenode *child, state_t state);
static state_t analyze_node(treenode *node, state_t state);
static char * state_str(state_t state);
static void init_var_state(scopetab_t *This);
static treenode *make_spec_call(treenode *init);
static void make_spec_header();
static treenode *make_spec_for_proc(treenode *test, treenode *incr, treenode *stemnt);
static treenode *make_spec_while_proc(treenode *test, treenode *stemnt);
static treenode *globalize(treenode *child);

/*  ###############################################################  */
leafnode *trans_leaf(leafnode *leaf)
{

  if (leaf == NULL){
    return(NULL);
  }

  switch(leaf->type) {

    case TN_IDENT:
/*    fprintf(fp,"Identifier %s\n", nmestr(leaf->data.sval)); */
      break;

		case TN_TYPE:
		case TN_INT:
		case TN_REAL:
		case TN_STRING:
		case TN_ELLIPSIS:
			break;

		default:
			fprintf(stderr, "Internal Error: Unknown type %d\n",
            leaf->type);
      break;
  }
  return(leaf);
}

/*  ###############################################################  */
static treenode *trans_node(treenode *node)
{

  leafnode *leaf;

  if (node == NULL){
    fprintf( stderr, "NULL node pointer 0x0\n");
    return(NULL);
  }

	if (node->type == TN_WHILE) {
	/*
	willey
	add trans_while here
	need to hack it out a little so it doesn't try ot trans the left and right
	after the case.
	*/
		node = (treenode *)trans_while((treenode *) node);
	} else {

  switch(node->type){

    case TN_EMPTY:

/* willey
	this is weird.  The decl is the prototype
	the definition is when we are really writing the function
	but we are changing the name in the decl and it works
*/
    case TN_FUNC_DEF:
printf ("I'm transing a def!\n");
      break;

    case TN_FUNC_DECL:
      leaf = (leafnode *)node->lnode;
      CurFunc = nmestr(leaf->data.sval);

			/*
				willey
				replace main --> mainX here
			*/
			/* willey
				if the function is "main", replace with "mainX"
			*/
/*			if (strncmp (CurFunc,"main",5) == 0) { */
printf ("at line %d, col %d, tok %d, name %s\n",node->line,node->col,node->tok,CurFunc);

			if (strcmp (CurFunc,"main") == 0) {
				/* willey
					install symbol "mainX" in the symbol table
					we now assign a new symbol to the leaf
					we lose the old symbol, oh well
				*/
				if ((leaf->data.sval = nmelook ("mainX",0)) == NULL) {
					printf ("error, cannot create mainX symbol\n");
					exit (1);
				}
			}

      LoopNum = 0;
      break;

    case TN_FUNC_CALL:
      break;
    case TN_BLOCK:
      break;
    case TN_DECL:
    case TN_ARRAY_DECL:
    case TN_STEMNT_LIST:
    case TN_TRANS_LIST:
    case TN_DECL_LIST:
    case TN_EXPR_LIST:
    case TN_NAME_LIST:
    case TN_ENUM_LIST:
    case TN_FIELD_LIST:
    case TN_PARAM_LIST:
    case TN_IDENT_LIST:
    case TN_COMP_DECL:
    case TN_BIT_FIELD:
    case TN_PNTR:
    /* Stores a list of type specifiers/qualifiers (int:const:etc) */
    case TN_TYPE_LIST:
    case TN_TYPE_NME:
    /* Stores initial values for arrays */
    case TN_INIT_LIST:
    case TN_INIT_BLK:
    case TN_OBJ_DEF:    /* Definition of struct: union: or enum */
    case TN_OBJ_REF:    /* Reference to struct: union: or enum */
    case TN_CAST:
    case TN_IF:
    case TN_ASSIGN:
    case TN_JUMP:
    case TN_FOR:

		/* willey
			this case is trapped above
		*/
    case TN_WHILE:
    case TN_DOWHILE:
    case TN_SWITCH:
    case TN_ELLIPSIS:
    case TN_LABEL:
    case TN_STEMNT:
    case TN_INDEX:     /* Index with [] */
    case TN_DEREF:     /* Dereference with * */
    case TN_SELECT:    /* -> and . */
    case TN_EXPR:
    case TN_COND_EXPR:
      break;
    case TN_IDENT:
      fprintf(stderr,"Unexpected identifier node 0x%0x\n",
              (unsigned int) node);
      break;

    case TN_TYPE:
      break;

    case TN_STRING:
      fprintf(stderr,"Unexpected string node 0x%0x\n", (unsigned int) node);
      break;

    case TN_INT:
      fprintf(stderr,"Unexpected integer node 0x%0x\n",
              (unsigned int) node);
      break;

    case TN_REAL:
      fprintf(stderr,"Unexpected real node 0x%0x\n", (unsigned int) node);
      break;

    default:
      fprintf(stderr,"Unknown node 0x%0x; internal error (type=%d)\n",
          (unsigned int) node, node->type);
      break;
  }

  node->lnode = trans_any(node->lnode);
  node->rnode = trans_any(node->rnode);

	} /* willey - end if not TN_WHILE block */

  return (node);

}

/*  ###############################################################  */
static if_node *trans_if(if_node *ifn)
{

  if (ifn == NULL){
      fprintf(stderr, "Unexpected Null If-node pointer 0x0\n");
      return(NULL);
  }


  ifn->cond = trans_any(ifn->cond);
  ifn->then_n = trans_any(ifn->then_n);
  ifn->else_n = trans_any(ifn->else_n);

  return(ifn);
}

/*  ###############################################################  */
static for_node *trans_for(for_node *forn)
{
  if_node *ifn;
	treenode *stmt;
	leafnode *leaf;
	 
  if (forn == NULL){
    fprintf(stderr, "Unexpected Null For-node pointer 0x0\n");
    return(NULL);
  }

  if (forn->type != TN_FUNC_DEF) {


    TrueLocals = GlobalLocals = NULL;
    init_var_state(ParseStack->contxt->syms->root);
    analyze_any(forn->test, ST_START);
    analyze_any(forn->incr, ST_INC);
    analyze_any(forn->stemnt, ST_START);

    sprintf(LoopName, "%s_%d", CurFunc, ++LoopNum);
    make_spec_header();

    stmt = make_node(TN_STEMNT);
    stmt->rnode = (treenode *)ifn = make_if (TN_IF);
    ifn->cond  = (treenode *) make_node(TN_FUNC_CALL);
    leaf =  make_leaf(TN_IDENT);
    leaf->data.sval = nmelook("spec_check", 10);
    ifn->cond->lnode = (treenode *)leaf;

    ifn->then_n = make_spec_call(forn->init);

		/* 
		willey
		initialize SpecProcs if this is the first procedure encountered
		*/

		if (SpecProcs == NULL) {
			SpecProcs = make_node(TN_TRANS_LIST);
			SpecProcs_L = SpecProcs;
		}

		SpecProcs_L->lnode = make_spec_for_proc(forn->test, forn->incr, forn->stemnt);
			SpecProcs_L->rnode = make_node(TN_TRANS_LIST);
		SpecProcs_L = SpecProcs_L->rnode;

		/*
		willey

		SpecProcs = make_spec_for_proc(forn->test, forn->incr, forn->stemnt);
		*/
		
		ifn->else_n = (treenode *) make_node(TN_STEMNT);
		ifn->else_n->rnode = make_node(TN_BLOCK);
		ifn->else_n->rnode->rnode = make_node(TN_STEMNT_LIST);
		ifn->else_n->rnode->rnode->lnode = make_node(TN_STEMNT);
		ifn->else_n->rnode->rnode->lnode->rnode = (treenode *)forn;

	} else
		stmt = (treenode *)forn;
    
	forn->init = trans_any(forn->init);
	forn->test = trans_any(forn->test);
	forn->incr = trans_any(forn->incr);
	forn->stemnt = trans_any(forn->stemnt);

	return((for_node *) stmt);
}

static treenode *trans_while(treenode *whilen)
{
  if_node *ifn;
	treenode *stmt;
	leafnode *leaf;
	
	 
  if (whilen == NULL){
    fprintf(stderr, "Unexpected Null while-node pointer 0x0\n");
    return(NULL);
  }

  if (whilen->type != TN_FUNC_DEF) {

    TrueLocals = GlobalLocals = NULL;
    init_var_state(ParseStack->contxt->syms->root);
    analyze_any(whilen->lnode, ST_START);
    analyze_any(whilen->rnode, ST_START);

    sprintf(LoopName, "%s_%d", CurFunc, ++LoopNum);
    make_spec_header();

    stmt = make_node(TN_STEMNT);
    stmt->rnode = (treenode *)ifn = make_if (TN_IF);
    ifn->cond  = (treenode *) make_node(TN_FUNC_CALL);
    leaf =  make_leaf(TN_IDENT);
    leaf->data.sval = nmelook("spec_check", 10);
    ifn->cond->lnode = (treenode *)leaf;

    ifn->then_n = make_spec_call((treenode *) NULL);

		/* 
		willey
		initialize SpecProcs if this is the first procedure encountered
		*/

		if (SpecProcs == NULL) {
			SpecProcs = make_node(TN_TRANS_LIST);
			SpecProcs_L = SpecProcs;
		}

		SpecProcs_L->lnode = make_spec_while_proc(whilen->lnode, whilen->rnode);
			SpecProcs_L->rnode = make_node(TN_TRANS_LIST);
		SpecProcs_L = SpecProcs_L->rnode;

		/*
		willey

		SpecProcs = make_spec_while_proc(whilen->lnode, whilen->rnode);
		*/
		
		ifn->else_n = (treenode *) make_node(TN_STEMNT);
		ifn->else_n->rnode = make_node(TN_BLOCK);
		ifn->else_n->rnode->rnode = make_node(TN_STEMNT_LIST);
		ifn->else_n->rnode->rnode->lnode = make_node(TN_STEMNT);
		ifn->else_n->rnode->rnode->lnode->rnode = (treenode *)whilen;

	} else
		stmt = (treenode *)whilen;
    
	whilen->lnode = trans_any(whilen->lnode);
	whilen->rnode = trans_any(whilen->rnode);

	return((treenode *) stmt);
}

/*  ###############################################################  */

static char str[512];

static treenode *make_spec_call(treenode *init)
{
  treenode *stmt = NULL, *root = NULL;
  leafnode *leaf;
  symentry_t *symptr;
  int len;
 

  /* this routine generates code backwards i.e. post-order 
     tree traversal */


  root = (treenode *) make_node(TN_STEMNT);
  root->rnode = make_node(TN_BLOCK);
  stmt = root->rnode->rnode = make_node(TN_STEMNT_LIST);
	
  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		stmt->lnode = make_node(TN_STEMNT_LIST); 
 		stmt = stmt->lnode;
		stmt->rnode = make_node(TN_STEMNT); 
		stmt->rnode->rnode = make_node(TN_ASSIGN);
		stmt->rnode->rnode->tok = EQ;
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = symptr->nme;
		stmt->rnode->rnode->lnode = (treenode *) leaf;
		/* willey
			put parens around member
		*/
		len = sprintf(str, "(%sV->%s)", LoopName, nmestr(leaf->data.sval)); /* the lazy way */
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = nmelook(str, len);
		stmt->rnode->rnode->rnode = (treenode *) leaf;
	 }
  
  stmt->lnode = make_node(TN_STEMNT_LIST); 
  stmt = stmt->lnode;
  stmt->rnode = make_node(TN_STEMNT); 
  stmt->rnode->rnode = make_node(TN_FUNC_CALL);
  leaf = make_leaf(TN_IDENT);
  leaf->data.sval = nmelook("spec_begin", 10);
  stmt->rnode->rnode->lnode = (treenode *)leaf;
  stmt->rnode->rnode->rnode = make_node(TN_EXPR_LIST);
  leaf = make_leaf(TN_IDENT);
  leaf->data.sval = nmelook(LoopName, strlen(LoopName));
  stmt->rnode->rnode->rnode->lnode = (treenode *)leaf;
  leaf = make_leaf(TN_IDENT);
  len = sprintf(str, "%sV", LoopName);
  leaf->data.sval = nmelook(str, len);
  stmt->rnode->rnode->rnode->rnode = (treenode *)leaf;

  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		stmt->lnode = make_node(TN_STEMNT_LIST); 
 		stmt = stmt->lnode;
		stmt->rnode = make_node(TN_STEMNT); 
		stmt->rnode->rnode = make_node(TN_ASSIGN);
		stmt->rnode->rnode->tok = EQ;
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = symptr->nme;
		stmt->rnode->rnode->rnode = (treenode *) leaf;
		/* willey
			put parens around member
		*/
		len = sprintf(str, "(%sV->%s)", LoopName, leaf->data.sval->str); /* the lazy way */
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = nmelook(str, len);
		stmt->rnode->rnode->lnode = (treenode *) leaf;
	 }
  
	/* willey
		a while loop has no init, thus, it is null
		for loops will use this feature, however
	*/
	if (init != NULL) {
  	stmt->lnode = make_node(TN_STEMNT_LIST); 
  	stmt = stmt->lnode;
  	stmt->rnode = make_node(TN_STEMNT); 
  	stmt->rnode->rnode = init;
  }

  return(root);
}

/*  ###############################################################  */
static void make_spec_header()
{
  symentry_t *symptr;

	/* willey
		create the struct
	*/
  fprintf(header_fp,"typedef struct %s_s\n{\n", LoopName);
  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		fputs("\t", header_fp);
		print_type(symptr->node, header_fp);
		fprintf(header_fp, "\t%s;\n", symptr->nme->str);
	 }
  fprintf(header_fp,"}\n%s_t;\nextern %s_t *%sV;\nint %s(%s_t *nonlocals);\n\n",
			 LoopName, LoopName, LoopName, LoopName, LoopName);

	/* willey
		malloc up space for the struct in main
	*/
  fprintf(startup_fp,"%sV = (%s_t *) malloc(sizeof(%s_t));\n", LoopName, LoopName, LoopName);

  fprintf(shutdown_fp,"free(%sV);\n", LoopName);

  fprintf(decl_fp,"%s_t *%sV;\n", LoopName, LoopName);
}

/*  ###############################################################  */
static treenode *make_spec_for_proc(treenode *test, treenode *incr, treenode *stemnt)

{

  for_node *proc;
  leafnode *leaf;
  int len;
  treenode *decl, *stmt, *newstmt;
  symentry_t *symptr;
  if_node *ifn;

  test = copy_tree(test);
  incr = copy_tree(incr);
  stemnt = copy_tree(stemnt);


  proc = make_for(TN_FUNC_DEF);
  proc->init = (treenode *) NULL; /* default to int */

  proc->test = make_node(TN_FUNC_DECL);
  leaf = make_leaf(TN_IDENT);

  leaf->data.sval = nmelook(LoopName, strlen(LoopName));
  proc->test->lnode = (treenode *) leaf;
  proc->test->rnode = make_node(TN_DECL);
  proc->test->rnode->lnode = make_node(TN_TYPE_LIST);

  leaf = make_leaf(TN_TYPE);
  len = sprintf(str, "%s_t", LoopName);
  leaf->data.sval = nmelook(str, len);
  leaf->tok = TYPEDEF_NAME;
  proc->test->rnode->lnode->lnode = (treenode *) leaf;
  proc->test->rnode->rnode = make_node(TN_DECL);
  proc->test->rnode->rnode->lnode = make_node(TN_PNTR);
  leaf = make_leaf(TN_IDENT);
  leaf->data.sval = nmelook("nonlocals", 9);
  proc->test->rnode->rnode->rnode = (treenode *) leaf;

  proc->incr = (treenode *) NULL;

  proc->stemnt = make_node(TN_BLOCK);
  decl = proc->stemnt;

  for (symptr = TrueLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		decl->lnode = make_node(TN_DECL_LIST); 
 		decl = decl->lnode;
		decl->rnode = make_node(TN_DECL); 
		decl->rnode->lnode = copy_type(symptr->node);
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = symptr->nme;
		decl->rnode->rnode = (treenode *) leaf;
	 }

  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		if (symptr->state == ST_INCWR) {
		  decl->lnode = make_node(TN_DECL_LIST); 
		  decl = decl->lnode;
		  decl->rnode = make_node(TN_DECL); 
		  decl->rnode->lnode = copy_type(symptr->node);
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = symptr->nme;
		  decl->rnode->rnode = (treenode *) leaf;
		}
	 }

  
  stmt = NULL;
  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		if (symptr->state == ST_INCWR) {
		  newstmt = make_node(TN_STEMNT_LIST); 
		  newstmt->lnode = stmt;
		  stmt = newstmt;
		  stmt->rnode = make_node(TN_STEMNT); 
		  stmt->rnode->rnode = make_node(TN_ASSIGN);
		  stmt->rnode->rnode->tok = EQ;
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = symptr->nme;
		  stmt->rnode->rnode->lnode = (treenode *) leaf;

			/* willey
				this should be replaced with a tree fragment and multiple symbol table
				entries... some day
			*/
		/* willey
			put parens around member
		*/
		  len = sprintf(str, "(nonlocals->%s)", nmestr(leaf->data.sval)); /* the lazy way */
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = nmelook(str, len);
		  stmt->rnode->rnode->rnode = (treenode *) leaf;
		}
	 }

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = globalize(incr);

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = (treenode *) ifn = make_if(TN_IF);
  ifn->cond = make_node(TN_EXPR);
  ifn->cond->tok = NOT;

	/* willey
		should we globalize the test?
	*/
  ifn->cond->rnode = test;
  ifn->then_n = make_node(TN_STEMNT); 
  ifn->then_n->rnode = make_node(TN_JUMP); 
  ifn->then_n->rnode->tok = RETURN;
  leaf = make_leaf(TN_INT);
  leaf->data.ival = 1;
  ifn->then_n->rnode->lnode = (treenode *)leaf;

  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
		if (symptr->state == ST_INCWR)
		  symptr->state = ST_ASSIGN;
	  
  
  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = globalize((treenode *) stemnt);

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = make_node(TN_JUMP); 
  stmt->rnode->rnode->tok = RETURN;
  leaf = make_leaf(TN_INT);
  leaf->data.ival = 0;
  stmt->rnode->rnode->lnode = (treenode *)leaf;

  proc->stemnt->rnode = stmt;
  return((treenode *)proc);
  
}

/*  ###############################################################  */
static treenode *make_spec_while_proc(treenode *test, treenode *stemnt)

{

  for_node *proc;
  leafnode *leaf;
  int len;
  treenode *decl, *stmt, *newstmt;
  symentry_t *symptr;
  if_node *ifn;

  test = copy_tree(test);
  stemnt = copy_tree(stemnt);


  proc = make_for(TN_FUNC_DEF);
  proc->init = (treenode *) NULL; /* default to int */

  proc->test = make_node(TN_FUNC_DECL);
  leaf = make_leaf(TN_IDENT);

  leaf->data.sval = nmelook(LoopName, strlen(LoopName));
  proc->test->lnode = (treenode *) leaf;
  proc->test->rnode = make_node(TN_DECL);
  proc->test->rnode->lnode = make_node(TN_TYPE_LIST);

  leaf = make_leaf(TN_TYPE);
  len = sprintf(str, "%s_t", LoopName);
  leaf->data.sval = nmelook(str, len);
  leaf->tok = TYPEDEF_NAME;
  proc->test->rnode->lnode->lnode = (treenode *) leaf;
  proc->test->rnode->rnode = make_node(TN_DECL);
  proc->test->rnode->rnode->lnode = make_node(TN_PNTR);
  leaf = make_leaf(TN_IDENT);
  leaf->data.sval = nmelook("nonlocals", 9);
  proc->test->rnode->rnode->rnode = (treenode *) leaf;

  proc->incr = (treenode *) NULL;

  proc->stemnt = make_node(TN_BLOCK);
  decl = proc->stemnt;

  for (symptr = TrueLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		decl->lnode = make_node(TN_DECL_LIST); 
 		decl = decl->lnode;
		decl->rnode = make_node(TN_DECL); 
		decl->rnode->lnode = copy_type(symptr->node);
		leaf = make_leaf(TN_IDENT);
		leaf->data.sval = symptr->nme;
		decl->rnode->rnode = (treenode *) leaf;
	 }

  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		if (symptr->state == ST_INCWR) {
		  decl->lnode = make_node(TN_DECL_LIST); 
		  decl = decl->lnode;
		  decl->rnode = make_node(TN_DECL); 
		  decl->rnode->lnode = copy_type(symptr->node);
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = symptr->nme;
		  decl->rnode->rnode = (treenode *) leaf;
		}
	 }

  
  stmt = NULL;
  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
	 {
		if (symptr->state == ST_INCWR) {
		  newstmt = make_node(TN_STEMNT_LIST); 
		  newstmt->lnode = stmt;
		  stmt = newstmt;
		  stmt->rnode = make_node(TN_STEMNT); 
		  stmt->rnode->rnode = make_node(TN_ASSIGN);
		  stmt->rnode->rnode->tok = EQ;
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = symptr->nme;
		  stmt->rnode->rnode->lnode = (treenode *) leaf;

			/* willey
				this should be replaced with a tree fragment and multiple symbol table
				entries... some day
			*/
		/* willey
			put parens around member
		*/
		  len = sprintf(str, "(nonlocals->%s)", nmestr(leaf->data.sval)); /* the lazy way */
		  leaf = make_leaf(TN_IDENT);
		  leaf->data.sval = nmelook(str, len);
		  stmt->rnode->rnode->rnode = (treenode *) leaf;
		}
	 }

/* willey
  while loops do not have this incr, so we do not add it

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = globalize(incr);
*/

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = (treenode *) ifn = make_if(TN_IF);

  ifn->cond = make_node(TN_EXPR);
  ifn->cond->tok = NOT;

	/* willey
		not sure if this is "correct" but I globalized test, not like in for loop
	*/
  ifn->cond->rnode = globalize(test);
  ifn->then_n = make_node(TN_STEMNT); 
  ifn->then_n->rnode = make_node(TN_JUMP); 
  ifn->then_n->rnode->tok = RETURN;
  leaf = make_leaf(TN_INT);
  leaf->data.ival = 1;
  ifn->then_n->rnode->lnode = (treenode *)leaf;

  for (symptr = GlobalLocals; symptr!= NULL; symptr = symptr->nxt_local)
		if (symptr->state == ST_INCWR)
		  symptr->state = ST_ASSIGN;
	  
  
  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = globalize((treenode *) stemnt);

  newstmt = make_node(TN_STEMNT_LIST); 
  newstmt->lnode = stmt;
  stmt = newstmt;
  stmt->rnode =  make_node(TN_STEMNT); 
  stmt->rnode->rnode = make_node(TN_JUMP); 
  stmt->rnode->rnode->tok = RETURN;
  leaf = make_leaf(TN_INT);
  leaf->data.ival = 0;
  stmt->rnode->rnode->lnode = (treenode *)leaf;

  proc->stemnt->rnode = stmt;
  return((treenode *)proc);
  
}

/*  ###############################################################  */
static treenode *globalize(treenode *child)
{
  for_node *forn;
  if_node *ifn;
  leafnode *leaf;
  symentry_t *sym;
  scopetab_t *scope;
  int len;

  

  if (!child){
  } else {
	 
	 switch (child->which){
	 case LEAF_T:
		leaf = (leafnode *) child;
		if (TN_IDENT == leaf->type) {

		  scope = (scopetab_t *)leaf->scp;
		  sym = scopetab_find(scope, leaf->data.sval);
		  
/* willey
	changed FILE_SCOPE to EXTERN_SCOPE to attempt to fix queen c bug
*/
		  if (is_vardecl(sym )&& sym->level > FILE_SCOPE && 
				sym->state != ST_ASSIGN)
			 {
		/* willey
			put parens around member
		*/
				len = sprintf(str, "(nonlocals->%s)", leaf->data.sval->str); /* the lazy way */
				leaf->data.sval = nmelook(str, len);
			 }
		}
		break;
		
	 case NODE_T:
		globalize (child->lnode);
		globalize (child->rnode);
		break;
		
	 case IF_T:
		ifn = (if_node *) child;
		globalize(ifn->cond);
		globalize(ifn->then_n);
		globalize(ifn->else_n);
		break;
		
	 case FOR_T:
		forn = (for_node *) child;
		globalize(forn->init);
		globalize(forn->test);
		globalize(forn->incr);
		globalize(forn->stemnt);
		break;
		
	 case NONE_T:
		fprintf(stderr, "Node 0x%0x: no-type\n", (unsigned int) child);
		break;
		
	 default:
		fprintf(stderr,
				  "Node 0x%0x internal error: unknown which-type %d\n",
				  (unsigned int) child, child->which);
		break;
	 }
  }
  return(child);
}


/*  ###############################################################  */
static treenode *trans_any(treenode *child)
{
	if (!child){
	} else {

		switch (child->which){
			case LEAF_T:
				child = (treenode *)trans_leaf((leafnode *) child);
				break;

			case NODE_T:
				child = trans_node(child);
				break;
				
			case IF_T:
				child = (treenode *)trans_if((if_node *) child);
				break;
				
			case FOR_T:
				child = (treenode *)trans_for((for_node *) child);
				break;
				
			case NONE_T:
				fprintf(stderr, "Node 0x%0x: no-type\n", (unsigned int) child);
				break;

			default:
				fprintf(stderr,
								"Node 0x%0x internal error: unknown which-type %d\n",
								(unsigned int) child, child->which);
				break;
		}
	}

	return(child);
}

/*  ###############################################################  */
treenode *trans_tree(treenode *root)
{
	/*
  willey
  now handled at the insertion point
  SpecProcs = make_node(TN_TRANS_LIST);
	*/

printf("This is a test - about to trans the tree.\n");
	/* willey
		needed to send code snippets to different files
	*/
	header_fp = fopen ("speculation.h","a");
	startup_fp = fopen ("hydracat_startup.h","a");
	shutdown_fp = fopen ("hydracat_shutdown.h","a");
	decl_fp = fopen ("hydracat_decl.h","a");

	return(trans_any(root));
}




/*  ###############################################################  */
static state_t analyze_node(treenode *node, state_t state)
{

  if (node == NULL){
    fprintf( stderr, "NULL node pointer 0x0\n");
    return(ST_START);
  }

  switch(node->type){

		case TN_EMPTY:
		case TN_FUNC_DEF:
 		case TN_FUNC_DECL:
 		case TN_FUNC_CALL:
		case TN_BLOCK:
		case TN_DECL:
		case TN_ARRAY_DECL:
		case TN_STEMNT_LIST:
		case TN_TRANS_LIST:
		case TN_DECL_LIST:
		case TN_EXPR_LIST:
		case TN_NAME_LIST:
		case TN_ENUM_LIST:
		case TN_FIELD_LIST:
		case TN_PARAM_LIST:
		case TN_IDENT_LIST:
		case TN_COMP_DECL:
		case TN_BIT_FIELD:
		case TN_PNTR:    /* Stores a list of type specifiers/qualifiers (int:const:etc) */
		case TN_TYPE_LIST:
		case TN_TYPE_NME: /* Stores initial values for arrays */
		case TN_INIT_LIST:
		case TN_INIT_BLK:
		case TN_OBJ_DEF:    /* Definition of struct: union: or enum */
		case TN_OBJ_REF:    /* Reference to struct: union: or enum */
		case TN_CAST:
		case TN_IF:
			break;
		case TN_ASSIGN:
			if (node->tok != EQ) break;
			state = analyze_any(node->rnode, state);
			if (state == ST_INC)
				state = analyze_any(node->lnode, ST_INCWR);
			else
				state = analyze_any(node->lnode, ST_ASSIGN);
			return(state);
			break;
		case TN_JUMP:
		case TN_FOR:
		case TN_WHILE:
		case TN_DOWHILE:
		case TN_SWITCH:
		case TN_ELLIPSIS:
		case TN_LABEL:
		case TN_STEMNT:
		case TN_INDEX:     /* Index with [] */
			break;
		case TN_DEREF:     /* Dereference with * */
			if (state == ST_ASSIGN)
				state = ST_PTRASGN;
			break;
		case TN_SELECT:    /* -> and . */
			break;
		case TN_EXPR:
			if (state == ST_INC && (node->tok == INCR || node->tok == DECR)) 
			state = ST_INCWR;
			break;
		case TN_COND_EXPR:
			 break;
		case TN_IDENT:
			fprintf(stderr,"Unexpected identifier node 0x%0x\n",
							(unsigned int) node);
			break;

		case TN_TYPE:
			break;

		case TN_STRING:
			fprintf(stderr,"Unexpected string node 0x%0x\n", (unsigned int) node);
			break;

		case TN_INT:
			fprintf(stderr,"Unexpected integer node 0x%0x\n",
							(unsigned int) node);
			break;

		case TN_REAL:
			fprintf(stderr,"Unexpected real node 0x%0x\n", (unsigned int) node);
			break;

		default:
			fprintf(stderr,"Unknown node 0x%0x; internal error (type=%d)\n",
						(unsigned int) node, node->type);
			break;
	}

	state = analyze_any(node->lnode, state);
	state = analyze_any(node->rnode, state);

	return (state);

}

/*  ###############################################################  */
static state_t analyze_any(treenode *child, state_t state)
{
	for_node *forn;
//	treenode *whilen;
	if_node *ifn;
	leafnode *leaf;
	symentry_t *sym;
	scopetab_t *scope;
  

	if (!child){
	} else {

		switch (child->which){
      case LEAF_T:
				leaf = (leafnode *) child;
				if (TN_IDENT == leaf->type) {
				scope = (scopetab_t *)leaf->scp;
				sym = scopetab_find(scope, leaf->data.sval);
			
				if (is_vardecl(sym))
				{
					if (sym->state == ST_UNDEF) 
					{
					  sym->state = state;
/* willey
	changed FILE_SCOPE to EXTERN_SCOPE to attempt to fix queen c bug
*/
					  if (sym->level > FILE_SCOPE){
							if (sym->state != ST_ASSIGN) {
								sym->nxt_local = GlobalLocals;
								GlobalLocals = sym;
							} else {
								sym->nxt_local = TrueLocals;
								TrueLocals = sym;
							}
						}
					} else if (state == ST_INCWR)
						sym->state = state;
/* here */
				  
/* willey
	I think this is where it screws up on 'c'
*/
/* willey
	changed FILE_SCOPE to EXTERN_SCOPE to attempt to fix queen c bug
*/
					 if (sym->level > FILE_SCOPE)
						printf("Local var %s %d %s %s\n", nmestr(leaf->data.sval), 
								 sym->level, state_str(state), state_str(sym->state));
					 else
						printf("Global var %s %d %s %s\n", nmestr(leaf->data.sval), 
								 sym->level, state_str(state), state_str(sym->state)); 

/* willey
this is the end of the disable comment
*/
				  }
			 }
			 state = ST_START;
			 break;
     
        case NODE_T:
					/* need to change for while node */
/*
willey
I think we do not do it here, wait for the translate part

					if (child->type == TN_WHILE) {
			 			whilen = (treenode *) child;
			 			state = analyze_any(whilen->lnode, state);
			 			state = analyze_any(whilen->rnode, state);
          } else {
*/
			 			state = analyze_node(child, state);
/*
willey

          }
*/
			 		break;
			 
        case IF_T:
					ifn = (if_node *) child;
					state = analyze_any(ifn->cond, state);
					state = analyze_any(ifn->then_n, state);
					state = analyze_any(ifn->else_n, state);
					break;
			 
        case FOR_T:
					forn = (for_node *) child;
					state = analyze_any(forn->init, state);
					state = analyze_any(forn->test, state);
					state = analyze_any(forn->incr, state);
					state = analyze_any(forn->stemnt, state);
					break;
			 
        case NONE_T:
					fprintf(stderr, "Node 0x%0x: no-type\n", (unsigned int) child);
					break;
			 
        default:
					fprintf(stderr,
						"Node 0x%0x internal error: unknown which-type %d\n",
						(unsigned int) child, child->which);
					break;
        }
		}
  
  return(state);
}


/* ERROR state_str defined but not used */
static char * state_str(state_t state)
{
 
  switch (state) {
  case ST_START:
	 return ("START");
	 break;
  case ST_ASSIGN:
	 return ("ASSIGN");
	 break;
  case ST_PTRASGN:
	 return ("PTRASGN");
	 break;
  case ST_UNDEF:
	 return ("UNDEF");
	 break;

  default:
	 return ("Invalid state");
	 break;

  }
}


/*  ###############################################################  */
static void init_var_state(scopetab_t *This)
{
  int j;
  symentry_t *list;

  if (This->htab){
	 for (j=0; j < This->htab->tsize; j++)
		{
		  for (list = This->htab->tab[j]; list; list=list->next)
			 {
				  list->state = ST_UNDEF;
				  list->nxt_local = NULL;
				
			 }
		}
  }
  
  for (j=0; j < This->nchild; j++)
    init_var_state(This->children[j]);
}












